//  (C) Copyright Vaclav Vesely.
//  Use, modification and distribution are subject to the Boost Software
//  License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
//  http://www.boost.org/LICENSE_1_0.txt).
//
//  See http://www.boost.org/libs/type_traits for most recent version including
//  documentation.

#ifndef BOOST_TT_IS_MOVABLE_HPP_INCLUDED
#define BOOST_TT_IS_MOVABLE_HPP_INCLUDED

#include <boost/config.hpp>
#include <boost/preprocessor/comma.hpp>
#include <boost/type_traits/has_trivial_copy.hpp>

// should be the last #include
#include <boost/type_traits/detail/bool_trait_def.hpp>

//-----------------------------------------------------------------------------

namespace boost {

    BOOST_TT_AUX_BOOL_TRAIT_DEF1(is_movable, Type,
        ::boost::has_trivial_copy<Type>::value)

} // namespace boost

//-----------------------------------------------------------------------------

#define BOOST_MOVABLE(type) \
    template<> \
    struct ::boost::is_movable<type>: \
        ::boost::integral_constant<bool, true> \
    { \
    }; \
/**/

#define BOOST_MOVABLE_1P(type, param1, value) \
    template<param1> \
    struct ::boost::is_movable<type<param1> >: \
        ::boost::integral_constant<bool, value> \
    { \
    }; \
/**/

#define BOOST_MOVABLE_1(type) \
    BOOST_MOVABLE_1P(type, \
        typename Param1, \
        true) \
/**/

#define BOOST_MOVABLE_2P(type, param1, param2, value) \
    template<param1, param2> \
    struct ::boost::is_movable<type<param1, param2> >: \
        ::boost::integral_constant<bool, value> \
    { \
    }; \
/**/

#define BOOST_MOVABLE_2(type) \
    BOOST_MOVABLE_2P(type, \
        typename Param1, typename Param2, \
        true) \
/**/

#define BOOST_MOVABLE_3P(type, param1, param2, param3, value) \
    template<param1, param2, param3> \
    struct ::boost::is_movable<type<param1, param2, param3> >: \
        ::boost::integral_constant<bool, value> \
    { \
    }; \
/**/

#define BOOST_MOVABLE_3(type) \
    BOOST_MOVABLE_3P(type, \
        typename Param1, typename Param2, typename Param3, \
        true) \
/**/

#define BOOST_MOVABLE_4P(type, param1, param2, param3, param4, value) \
    template<param1, param2, param3, param4> \
    struct ::boost::is_movable<type<param1, param2, param3, param4> >: \
        ::boost::integral_constant<bool, value> \
    { \
    }; \
/**/

#define BOOST_MOVABLE_4(type) \
    BOOST_MOVABLE_4P(type, \
        typename Param1, typename Param2, typename Param3, typename Param4, \
        true) \
/**/

//-----------------------------------------------------------------------------

#if defined(BOOST_HAS_MOVABLE_STL)

namespace std {

    template<typename CharType, typename Traits, typename Allocator>
    class basic_string;

    template<typename Type>
    class complex;

    template<typename Type, typename Allocator>
    class deque;

    template<typename Type, typename Allocator>
    class list;

    template<typename Key, typename Type, typename Allocator, typename Compare>
    class map;

    template<typename Key, typename Type, typename Allocator, typename Compare>
    class multimap;

    template<typename Key, typename Allocator, typename Compare>
    class multiset;

    template<typename Type, typename Container, typename Compare>
    class priority_queue;

    template<typename Key, typename Allocator, typename Compare>
    class set;

    template<typename Type, typename Container>
    class stack;

    template<typename Type, typename Allocator>
    class vector;

    template<typename Type, typename Container>
    class queue;

} // namespace std

namespace boost {

    #if !defined(BOOST_NO_TEMPLATE_PARTIAL_SPECIALIZATION)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_3(
            typename CharType, typename Traits, typename Allocator,
            is_movable, std::basic_string<CharType BOOST_PP_COMMA() Traits
            BOOST_PP_COMMA() Allocator>, true)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_1(
            typename Type,
            is_movable, std::complex<Type>, true)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_2(
            typename Type, typename Allocator,
            is_movable, std::deque<Type BOOST_PP_COMMA() Allocator>, true)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_2(
            typename Type, typename Allocator,
            is_movable, std::list<Type BOOST_PP_COMMA() Allocator>, true)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_4(
            typename Key, typename Type, typename Allocator, typename Compare,
            is_movable, std::map<Key BOOST_PP_COMMA() Type BOOST_PP_COMMA()
            Allocator BOOST_PP_COMMA() Compare>, true)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_4(
            typename Key, typename Type, typename Allocator, typename Compare,
            is_movable, std::multimap<Key BOOST_PP_COMMA() Type
            BOOST_PP_COMMA() Allocator BOOST_PP_COMMA() Compare>, true)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_3(
            typename Key, typename Allocator, typename Compare,
            is_movable, std::multiset<Key BOOST_PP_COMMA() Allocator
            BOOST_PP_COMMA() Compare>, true)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_3(
            typename Type, typename Container, typename Compare,
            is_movable, std::priority_queue<Type BOOST_PP_COMMA()
            Container BOOST_PP_COMMA() Compare>, is_movable<Container>::value)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_3(
            typename Key, typename Allocator, typename Compare,
            is_movable, std::set<Key BOOST_PP_COMMA() Allocator
            BOOST_PP_COMMA() Compare>, true)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_2(
            typename Type, typename Container,
            is_movable, std::stack<Type BOOST_PP_COMMA() Container>,
            is_movable<Container>::value)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_2(
            typename Key, typename Allocator,
            is_movable, std::vector<Key BOOST_PP_COMMA() Allocator>, true)

        BOOST_TT_AUX_BOOL_TRAIT_PARTIAL_C_SPEC1_2(
            typename Type, typename Container,
            is_movable, std::queue<Type BOOST_PP_COMMA() Container>,
            is_movable<Container>::value)

    #endif // !defined(BOOST_NO_TEMPLATE_PARTIAL_SPECIALIZATION)

} // namespace boost

#endif //defined(BOOST_HAS_MOVABLE_STL)

//-----------------------------------------------------------------------------

#include <boost/type_traits/detail/bool_trait_undef.hpp>

#endif // BOOST_TT_IS_MOVABLE_HPP_INCLUDED
