//  (C) Copyright John Maddock, Vaclav Vesely. 
//  Use, modification and distribution are subject to the 
//  Boost Software License, Version 1.0. (See accompanying file 
//  LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)

#include "test.hpp"
#include "check_integral_constant.hpp"
#ifdef TEST_STD
#  include <type_traits>
#else
#  include <boost/type_traits/is_movable.hpp>
#endif
#include <complex>
#include <deque>
#include <list>
#include <map>
#include <queue>
#include <set>
#include <stack>
#include <string>
#include <vector>

class nonmovable_container:
    public std::vector<int>
{};

template<> 
struct boost::is_movable<nonmovable_container>:
    public boost::false_type
{};

TT_TEST_BEGIN(is_movable)

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<bool>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<bool const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<bool volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<bool const volatile>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<signed char>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<signed char const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<signed char volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<signed char const volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned char>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<char>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned char const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<char const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned char volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<char volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned char const volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<char const volatile>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned short>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<short>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned short const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<short const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned short volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<short volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned short const volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<short const volatile>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned int>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<int>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned int const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<int const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned int volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<int volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned int const volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<int const volatile>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned long>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<long>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned long const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<long const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned long volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<long volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned long const volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<long const volatile>::value, false);

#ifdef BOOST_HAS_LONG_LONG

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable< ::boost::ulong_long_type>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable< ::boost::long_long_type>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable< ::boost::ulong_long_type const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable< ::boost::long_long_type const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable< ::boost::ulong_long_type volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable< ::boost::long_long_type volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable< ::boost::ulong_long_type const volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable< ::boost::long_long_type const volatile>::value, false);

#endif

#ifdef BOOST_HAS_MS_INT64

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int8>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int8>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int8 const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int8 const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int8 volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int8 volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int8 const volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int8 const volatile>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int16>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int16>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int16 const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int16 const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int16 volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int16 volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int16 const volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int16 const volatile>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int32>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int32>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int32 const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int32 const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int32 volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int32 volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int32 const volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int32 const volatile>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int64>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int64>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int64 const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int64 const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int64 volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int64 volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<unsigned __int64 const volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<__int64 const volatile>::value, false);

#endif

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<float>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<float const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<float volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<float const volatile>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<double>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<double const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<double volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<double const volatile>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<long double>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<long double const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<long double volatile>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<long double const volatile>::value, false);


BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<int>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<void*>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<int*const>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<f1>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<f2>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<f3>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<mf1>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<mf2>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<mf3>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<mp>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<cmf>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<enum_UDT>::value, true);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<int&>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const int&>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<int[2]>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<int[3][2]>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<int[2][4][5][6][3]>::value, true);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<UDT>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<empty_UDT>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<void>::value, true);
// cases we would like to succeed but can't implement in the language:
BOOST_CHECK_SOFT_INTEGRAL_CONSTANT(::tt::is_movable<empty_POD_UDT>::value, true, false);
BOOST_CHECK_SOFT_INTEGRAL_CONSTANT(::tt::is_movable<POD_UDT>::value, true, false);
BOOST_CHECK_SOFT_INTEGRAL_CONSTANT(::tt::is_movable<POD_union_UDT>::value, true, false);
BOOST_CHECK_SOFT_INTEGRAL_CONSTANT(::tt::is_movable<empty_POD_union_UDT>::value, true, false);

bool movable_stl = false;

#if defined(BOOST_HAS_MOVABLE_STL) \
        && !defined(BOOST_NO_TEMPLATE_PARTIAL_SPECIALIZATION)
    movable_stl = true;
#endif

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<std::string>::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const std::string>::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::string>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::string>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<std::complex<float> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const std::complex<float> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::complex<float> >::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::complex<float> >::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<std::list<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const std::list<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::list<int> >::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::list<int> >::value, false);

typedef std::map<int, int> map_type;
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<map_type>::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const map_type>::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const map_type>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const map_type>::value, false);

typedef std::multimap<int, int> multimap_type;
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<multimap_type>::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const multimap_type>::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const multimap_type>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const multimap_type>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<std::multiset<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const std::multiset<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::multiset<int> >::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::multiset<int> >::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<std::priority_queue<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const std::priority_queue<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::priority_queue<int> >::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::priority_queue<int> >::value, false);

typedef std::priority_queue<int, nonmovable_container> nonmovable_priority_queue;
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<nonmovable_priority_queue>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const nonmovable_priority_queue>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile nonmovable_priority_queue>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const nonmovable_priority_queue>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<std::set<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const std::set<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::set<int> >::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::set<int> >::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<std::stack<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const std::stack<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::stack<int> >::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::stack<int> >::value, false);

typedef std::stack<int, nonmovable_container> nonmovable_stack;
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<nonmovable_stack>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const nonmovable_stack>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile nonmovable_stack>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const nonmovable_stack>::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<std::vector<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const std::vector<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::vector<int> >::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::vector<int> >::value, false);

BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<std::queue<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const std::queue<int> >::value, movable_stl);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::queue<int> >::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const std::queue<int> >::value, false);

typedef std::queue<int, nonmovable_container> nonmovable_queue;
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<nonmovable_queue>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<const nonmovable_queue>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile nonmovable_queue>::value, false);
BOOST_CHECK_INTEGRAL_CONSTANT(::tt::is_movable<volatile const nonmovable_queue>::value, false);

TT_TEST_END
